from __future__ import division
import math
import numpy as np
#maiken edit
import rospy
from custom_msgs.msg import ThreeDofForce


class MilliampereModel:
	def __init__(self, parameters):
		self.m_11 = parameters.get('m_11')
		self.m_12 = parameters.get('m_12', 0)
		self.m_13 = parameters.get('m_13', 0)
		self.m_21 = parameters.get('m_21', 0)
		self.m_22 = parameters.get('m_22')
		self.m_23 = parameters.get('m_23', 0)
		self.m_31 = parameters.get('m_31', 0)
		self.m_32 = parameters.get('m_32', 0)
		self.m_33 = parameters.get('m_33')
		self.X_u = parameters.get('X_u', 0)
		self.X_auu = parameters.get('X_auu', 0)
		self.X_uuu = parameters.get('X_uuu', 0)
		self.X_v = parameters.get('X_v', 0)
		self.X_r = parameters.get('X_r', 0)
		self.Y_u = parameters.get('Y_u', 0)
		self.Y_v = parameters.get('Y_v', 0)
		self.Y_avv = parameters.get('Y_avv', 0)
		self.Y_vvv = parameters.get('Y_vvv', 0)
		self.Y_arv = parameters.get('Y_arv', 0)
		self.Y_r = parameters.get('Y_r', 0)
		self.Y_avr = parameters.get('Y_avr', 0)
		self.Y_arr = parameters.get('Y_arr', 0)
		self.N_u = parameters.get('N_u', 0)
		self.N_v = parameters.get('N_v', 0)
		self.N_avv = parameters.get('N_avv', 0)
		self.N_arv = parameters.get('N_arv', 0)
		self.N_r = parameters.get('N_r', 0)
		self.N_arr = parameters.get('N_arr', 0)
		self.N_rrr = parameters.get('N_rrr', 0)
		self.N_avr = parameters.get('N_avr', 0)
		self.Y_ur = parameters.get('Y_ur', 0)
		self.N_ur = parameters.get('N_ur', 0)
		self.N_uv = parameters.get('N_uv', 0)
		self.use_coriolis = parameters.get('use_coriolis', True)
		self.l_1 = parameters.get('l_1')
		self.l_2 = parameters.get('l_2')
		self.thruster_force_polynomial = np.array(parameters.get('thruster_force_polynomial'))

		self.M = np.array([
			[self.m_11, self.m_12, self.m_13],
			[self.m_21, self.m_22, self.m_23],
			[self.m_31, self.m_32, self.m_33]
		], dtype='float')

		self.M_inv = np.linalg.inv(self.M)

	
	def coriolis(self, nu):
		u = nu[0]
		v = nu[1]
		r = nu[2]

		# c_13 = -self.m_21 * u - self.m_22 * v - self.m_23 * r
		# c_23 = self.m_11 * u + self.m_12 * v + self.m_13 * r
		c_13 = -0.5 * (self.m_21 + self.m_12) * u - self.m_22 * v - 0.5 * (self.m_23 + self.m_32) * r
		c_23 =  self.m_11 * u + 0.5 * (self.m_21 + self.m_12) * v + 0.5 * (self.m_13 + self.m_31) * r
		c_31 = -c_13
		c_32 = -c_23

		C = np.array([
			[0, 0, c_13],
			[0, 0, c_23],
			[c_31, c_32, 0]
		])

		return C
	
	def damping(self, nu):
		u = nu[0]
		v = nu[1]
		r = nu[2]

		d_11 = -self.X_u - self.X_auu * np.fabs(u) - self.X_uuu * u**2
		d_12 = -self.X_v
		d_13 = -self.X_r
		d_21 = -self.Y_u
		d_22 = -self.Y_v - self.Y_avv * np.fabs(v) - self.Y_arv * np.fabs(r) - self.Y_vvv * v**2
		d_23 = -self.Y_r - self.Y_avr * np.fabs(v) - self.Y_arr * np.fabs(r) - self.Y_ur * u
		d_31 = -self.N_u
		d_32 = -self.N_v - self.N_avv * np.fabs(v) - self.N_arv * np.fabs(r) - self.N_uv * u
		d_33 = -self.N_r - self.N_avr * np.fabs(v) - self.N_arr * np.fabs(r) - self.N_ur * u

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D
	
	def coriolis_wrt_psi(self,nu,V_c,psi):
		u = nu[0]
		v = nu[1]
		r = nu[2]
		v_cx = V_c[0]
		v_cy = V_c[1]

		c_13 = -0.5 * (self.m_21 + self.m_12)*(v_cx*np.sin(psi)-v_cy*np.cos(psi)) - self.m_22*(v_cx*np.cos(psi) + v_cy*np.sin(psi)) 
		c_23 =  self.m_11*(v_cx*np.sin(psi)-v_cy*np.cos(psi)) + 0.5 * (self.m_21 + self.m_12)*(v_cx*np.cos(psi) + v_cy*np.sin(psi))
		c_31 = -c_13
		c_32 = -c_23

		C = np.array([
			[0, 0, c_13],
			[0, 0, c_23],
			[c_31, c_32, 0]
		])

		return C

	def coriolis_wrt_u(self):
		c_13 = -0.5 * (self.m_21 + self.m_12)
		c_23 = self.m_11
		c_31 = -c_13
		c_32 = -c_23

		C = np.array([
			[0, 0, c_13],
			[0, 0, c_23],
			[c_31, c_32, 0]
		])

		return C

	def coriolis_wrt_v(self):
		c_13 = - self.m_22 
		c_23 = 0.5 * (self.m_21 + self.m_12)
		c_31 = -c_13
		c_32 = -c_23

		C = np.array([
			[0, 0, c_13],
			[0, 0, c_23],
			[c_31, c_32, 0]
		])

		return C

	def coriolis_wrt_r(self):
		c_13 = - 0.5 * (self.m_23 + self.m_32)
		c_23 = 0.5 * (self.m_13 + self.m_31)
		c_31 = -c_13
		c_32 = -c_23

		C = np.array([
			[0, 0, c_13],
			[0, 0, c_23],
			[c_31, c_32, 0]
		])

		return C

	def coriolis_wrt_vcx(self, psi):

		c_13 = -0.5 * (self.m_21 + self.m_12)*(-np.cos(psi)) - self.m_22*np.sin(psi)
		c_23 =  self.m_11*(-np.cos(psi)) + 0.5 * (self.m_21 + self.m_12)*np.sin(psi)
		c_31 = -c_13
		c_32 = -c_23

		C = np.array([
			[0, 0, c_13],
			[0, 0, c_23],
			[c_31, c_32, 0]
		])

		return C

	def coriolis_wrt_vcy(self, psi):

		c_13 = -0.5 * (self.m_21 + self.m_12)*(-np.sin(psi)) - self.m_22*(-np.cos(psi))
		c_23 =  self.m_11*(-np.sin(psi)) + 0.5 * (self.m_21 + self.m_12)*(-np.cos(psi))
		c_31 = -c_13
		c_32 = -c_23

		C = np.array([
			[0, 0, c_13],
			[0, 0, c_23],
			[c_31, c_32, 0]
		])

		return C

	def damping(self, nu):
		u = nu[0]
		v = nu[1]
		r = nu[2]

		d_11 = -self.X_u - self.X_auu * np.fabs(u) - self.X_uuu * u**2
		d_12 = -self.X_v
		d_13 = -self.X_r
		d_21 = -self.Y_u
		d_22 = -self.Y_v - self.Y_avv * np.fabs(v) - self.Y_arv * np.fabs(r) - self.Y_vvv * v**2
		d_23 = -self.Y_r - self.Y_avr * np.fabs(v) - self.Y_arr * np.fabs(r) - self.Y_ur * u
		d_31 = -self.N_u
		d_32 = -self.N_v - self.N_avv * np.fabs(v) - self.N_arv * np.fabs(r) - self.N_uv * u
		d_33 = -self.N_r - self.N_avr * np.fabs(v) - self.N_arr * np.fabs(r) - self.N_ur * u

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D



	def damping_wrt_psi(self, nu, V_c, psi):
		# u_r = u - V_cx*cos(psi) - V_cy*sin(psi)
		# v_r = v + V_cx*sin(psi) - V_Cy*cos(psi)
		# r_r = r
		u = nu[0]
		v = nu[1]
		r = nu[2]
		v_cx = V_c[0]
		v_cy = V_c[1]

		u_r = u - v_cx*np.cos(psi) - v_cy*np.sin(psi)
		v_r = v + v_cx*np.sin(psi) - v_cy*np.cos(psi)
		
		u_r_wrt_psi = v_cx*np.sin(psi) -v_cy*np.cos(psi)
		v_r_wrt_psi = v_cx*np.cos(psi) + v_cy*np.sin(psi)

		u_r_squared_wrt_psi = 2*u*(v_cx*np.sin(psi)-v_cy*np.cos(psi))- (v_cx**2)*np.sin(2*psi) \
			+ (v_cy**2)*(np.sin(2*psi)) + 2*v_cx*v_cy*np.cos(2*psi) 
		v_r_squared_wrt_psi = 2*v*(v_cx*np.cos(psi) + v_cy*np.sin(psi)) + (v_cx**2)*np.sin(2*psi) \
			- (v_cy**2)*np.sin(2*psi) - 2*v_cx*v_cy*np.cos(2*psi)
			
		# |f(x)|' = (f(x)/|f(x)|)*f'(x)
		# (f(x)/|f(x)|) = 1 if f(x)>0, (f(x)/|f(x)|) = -1 if f(x)<0
		# if f(x) = 0 --> 0.0000001 
		# abs(r) = -1 if r<0, 1 if r>0
		if u_r > 0:
			sign_ur = 1
		elif u_r < 0: 
			sign_ur = -1
		else:
			sign_ur = 0.000001

		if v_r > 0:
			sign_vr = 1
		elif v_r < 0: 
			sign_vr = -1
		else:
			sign_vr = 0.000001

		abs_ur_wrt_psi = sign_ur*(v_cx*np.sin(psi) - v_cy*np.cos(psi))
		abs_vr_wrt_psi = sign_vr*(v_cx*np.cos(psi) + v_cy*np.sin(psi))

		d_11 = - self.X_auu*abs_ur_wrt_psi - self.X_uuu*u_r_squared_wrt_psi 
		d_12 = 0.0
		d_13 = 0.0
		d_21 = 0.0
		d_22 = - self.Y_avv*abs_vr_wrt_psi - self.Y_vvv*v_r_squared_wrt_psi
		d_23 = - self.Y_avr*abs_vr_wrt_psi - self.Y_ur*u_r_wrt_psi
		d_31 = 0.0
		d_32 = - self.N_avv*abs_vr_wrt_psi - self.N_uv*u_r_wrt_psi
		d_33 = - self.N_avr*abs_vr_wrt_psi - self.N_ur*u_r_wrt_psi

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D

	def damping_wrt_u(self, nu, V_c, psi):
		# u_r = u - V_cx*cos(psi) - V_cy*sin(psi)
		# v_r = v + V_cx*sin(psi) - V_Cy*cos(psi)
		# r_r = r
		u = nu[0]
		v = nu[1]
		r = nu[2]
		v_cx = V_c[0]
		v_cy = V_c[1]
		u_r = u - v_cx*np.cos(psi) - v_cy*np.sin(psi)
		v_r = v + v_cx*np.sin(psi) - v_cy*np.cos(psi)
		u_r_wrt_u = 1.0
		u_r_squared = u**2 -2*u*(v_cx*np.cos(psi) + v_cy*np.sin(psi)) + (v_cx**2)*(np.cos(psi)**2) \
			+ (v_cy**2)*(np.sin(psi)**2) + 2*v_cx*v_cy*np.cos(psi)*np.sin(psi)
		# d/dpsi cos^2(psi) = -sin(2*psi)
		# d/dpsi sin^2(psi) = sin(2*psi)
		u_r_squared_wrt_u = 2*u - 2*(v_cx*np.cos(psi) + v_cy*np.sin(psi))


		# |f(x)|' = (f(x)/|f(x)|)*f'(x)
		# (f(x)/|f(x)|) = 1 if f(x)>0, (f(x)/|f(x)|) = -1 if f(x)<0
		# if f(x) = 0 --> 0.0000001 
		# abs(r) = -1 if r<0, 1 if r>0
		if u_r > 0:
			sign_ur = 1
		elif u_r < 0: 
			sign_ur = -1
		else:
			sign_ur = 0.000001

		abs_ur_wrt_u = sign_ur*1.0

		d_11 = - self.X_auu*abs_ur_wrt_u - self.X_uuu*u_r_squared_wrt_u
		d_12 = 0.0
		d_13 = 0.0
		d_21 = 0.0
		d_22 = 0.0
		d_23 = - self.Y_ur*u_r_wrt_u
		d_31 = 0.0
		d_32 = - self.N_uv*u_r_wrt_u
		d_33 = - self.N_ur*u_r_wrt_u

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D

	def damping_wrt_v(self, nu, V_c, psi):
		# u_r = u - V_cx*cos(psi) - V_cy*sin(psi)
		# v_r = v + V_cx*sin(psi) - V_Cy*cos(psi)
		# r_r = r
		u = nu[0]
		v = nu[1]
		r = nu[2]
		v_cx = V_c[0]
		v_cy = V_c[1]
		u_r = u - v_cx*np.cos(psi) - v_cy*np.sin(psi)
		v_r = v + v_cx*np.sin(psi) - v_cy*np.cos(psi)

		v_r_wrt_v = 1.0

		v_r_squared = v**2 + 2*v*(v_cx*np.sin(psi) - v_cy*np.cos(psi)) + (v_cx**2)*(np.sin(psi)**2) \
			+ (v_cy**2)*(np.cos(psi)**2) - 2*v_cx*v_cy*np.cos(psi)*np.sin(psi)

		v_r_squared_wrt_v = 2*v + 2*(v_cx*np.sin(psi) - v_cy*np.cos(psi))

		# |f(x)|' = (f(x)/|f(x)|)*f'(x)
		# (f(x)/|f(x)|) = 1 if f(x)>0, (f(x)/|f(x)|) = -1 if f(x)<0
		# if f(x) = 0 --> 0.0000001 
		# abs(r) = -1 if r<0, 1 if r>0

		if v_r > 0:
			sign_vr = 1
		elif v_r < 0: 
			sign_vr = -1
		else:
			sign_vr = 0.000001

		abs_vr_wrt_v = sign_vr

		d_11 = 0.0
		d_12 = 0.0
		d_13 = 0.0
		d_21 = 0.0
		d_22 = - self.Y_avv*abs_vr_wrt_v - self.Y_vvv*v_r_squared_wrt_v
		d_23 = - self.Y_avr*abs_vr_wrt_v
		d_31 = 0.0
		d_32 = - self.N_avv*abs_vr_wrt_v
		d_33 = - self.N_avr*abs_vr_wrt_v

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D

	def damping_wrt_r(self, nu):
		u = nu[0]
		v = nu[1]
		r = nu[2]

		if r > 0:
			sign_r = 1
		elif r < 0: 
			sign_r = -1
		else:
			sign_r = 0.000001

		d_11 = 0.0
		d_12 = 0.0
		d_13 = 0.0
		d_21 = 0.0
		d_22 = - self.Y_arv*sign_r 
		d_23 = - self.Y_arr*sign_r
		d_31 = 0.0
		d_32 = - self.N_arv*sign_r
		d_33 = - self.N_arr*sign_r

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D

	def damping_wrt_vcx(self, nu, V_c, psi):
		# u_r = u - V_cx*cos(psi) - V_cy*sin(psi)
		# v_r = v + V_cx*sin(psi) - V_Cy*cos(psi)
		# r_r = r
		u = nu[0]
		v = nu[1]
		r = nu[2]
		v_cx = V_c[0]
		v_cy = V_c[1]
		u_r = u - v_cx*np.cos(psi) - v_cy*np.sin(psi)
		v_r = v + v_cx*np.sin(psi) - v_cy*np.cos(psi)

		u_r_wrt_v_cx = -np.cos(psi)
		v_r_wrt_v_cx = np.sin(psi)

		# d/dpsi cos^2(psi) = -sin(2*psi)
		# d/dpsi sin^2(psi) = sin(2*psi)
		u_r_squared_wrt_v_cx = -2*u*np.cos(psi) + 2*v_cx*(np.cos(psi)**2) + 2*v_cy*np.cos(psi)*np.sin(psi)
		v_r_squared_wrt_v_cx = 2*v*np.sin(psi) + 2*v_cx*(np.sin(psi)**2)- 2*v_cy*np.cos(psi)*np.sin(psi)


		# |f(x)|' = (f(x)/|f(x)|)*f'(x)
		# (f(x)/|f(x)|) = 1 if f(x)>0, (f(x)/|f(x)|) = -1 if f(x)<0
		# if f(x) = 0 --> 0.0000001 
		# abs(r) = -1 if r<0, 1 if r>0
		if u_r > 0:
			sign_ur = 1
		elif u_r < 0: 
			sign_ur = -1
		else:
			sign_ur = 0.000001

		if v_r > 0:
			sign_vr = 1
		elif v_r < 0: 
			sign_vr = -1
		else:
			sign_vr = 0.000001

		abs_ur_wrt_v_cx = sign_ur*(-np.cos(psi))
		abs_vr_wrt_v_cx = sign_vr*np.sin(psi)

			

		d_11 = - self.X_auu * abs_ur_wrt_v_cx - self.X_uuu*u_r_squared_wrt_v_cx
		d_12 = 0.0
		d_13 = 0.0
		d_21 = 0.0
		d_22 = - self.Y_avv * abs_vr_wrt_v_cx - self.Y_vvv*v_r_squared_wrt_v_cx
		d_23 = - self.Y_avr * abs_vr_wrt_v_cx - self.Y_ur*u_r_wrt_v_cx
		d_31 = 0.0
		d_32 = - self.N_avv * abs_vr_wrt_v_cx - self.N_uv*u_r_wrt_v_cx
		d_33 = - self.N_avr * abs_vr_wrt_v_cx - self.N_ur*u_r_wrt_v_cx

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D


	def damping_wrt_vcy(self, nu, V_c, psi):
		# u_r = u - V_cx*cos(psi) - V_cy*sin(psi)
		# v_r = v + V_cx*sin(psi) - V_Cy*cos(psi)
		# r_r = r
		u = nu[0]
		v = nu[1]
		r = nu[2]
		v_cx = V_c[0]
		v_cy = V_c[1]
		u_r = u - v_cx*np.cos(psi) - v_cy*np.sin(psi)
		v_r = v + v_cx*np.sin(psi) - v_cy*np.cos(psi)

		u_r_wrt_v_cy = -np.sin(psi)
		v_r_wrt_v_cy = -np.cos(psi)

		u_r_squared_wrt_v_cy = -2*u*np.sin(psi) + 2*v_cy*(np.sin(psi)**2) + 2*v_cx*np.cos(psi)*np.sin(psi)
		v_r_squared_wrt_v_cy = -2*v*np.cos(psi) + 2*v_cy*(np.cos(psi)**2) - 2*v_cx*np.cos(psi)*np.sin(psi)

		# |f(x)|' = (f(x)/|f(x)|)*f'(x)
		# (f(x)/|f(x)|) = 1 if f(x)>0, (f(x)/|f(x)|) = -1 if f(x)<0
		# if f(x) = 0 --> 0.0000001 
		# abs(r) = -1 if r<0, 1 if r>0
		if u_r > 0:
			sign_ur = 1
		elif u_r < 0: 
			sign_ur = -1
		else:
			sign_ur = 0.000001

		if v_r > 0:
			sign_vr = 1
		elif v_r < 0: 
			sign_vr = -1
		else:
			sign_vr = 0.000001

		abs_ur_wrt_v_cy = sign_ur*(-np.sin(psi))
		abs_vr_wrt_v_cy = sign_vr*(-np.cos(psi))
			

		d_11 = - self.X_auu*abs_ur_wrt_v_cy  - self.X_uuu*u_r_squared_wrt_v_cy 
		d_12 = 0.0
		d_13 = 0.0
		d_21 = 0.0
		d_22 = - self.Y_avv*abs_vr_wrt_v_cy - self.Y_vvv *v_r_squared_wrt_v_cy
		d_23 = - self.Y_avr*abs_vr_wrt_v_cy - self.Y_ur*u_r_wrt_v_cy
		d_31 = 0.0
		d_32 = - self.N_avv*abs_vr_wrt_v_cy - self.N_uv*u_r_wrt_v_cy
		d_33 = - self.N_avr*abs_vr_wrt_v_cy - self.N_ur*u_r_wrt_v_cy

		D = np.array([
			[d_11, d_12, d_13],
			[d_21, d_22, d_23],
			[d_31, d_32, d_33]
		])

		return D
